﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace ConcurrencyAndArrays
{
    class Histogram
    {
        protected long sampleCount;
        protected ushort[] sampleValueCounts;

        public virtual long SampleCount { get { return sampleCount; } }
        public ushort[] SampleValueCounts { get { return sampleValueCounts; } }

        public Histogram(int range)
        {
            // + 1 to allow 0 - range, inclusive
            this.sampleValueCounts = new ushort[range + 1];
        }

        public virtual void AddSample(int value)
        {
            ++sampleValueCounts[value];
            ++sampleCount;
        }

        public virtual void AddSample(int value, int threadId)
        {
            // This will be overridden in a class below to use the threadId.
            this.AddSample(value);
        }
    }

    class HistogramLocked : Histogram
    {
        private object addLock = new object();

        public HistogramLocked(int range) : base(range) { }

        public override void AddSample(int value)
        {
            lock (addLock)
            {
                ++sampleValueCounts[value];
                ++sampleCount;
            }
        }
    }

    class HistogramInterlocked : Histogram
    {
        public HistogramInterlocked(int range) : base(range) { }

        public override void AddSample(int value)
        {
            ++sampleValueCounts[value];
            Interlocked.Increment(ref this.sampleCount);
        }
    }

    class HistogramThreadBuckets : Histogram
    {
        private long[] valueBuckets;

        public long[] ValueBuckets { get { return this.valueBuckets; } }

        public override long SampleCount
        {
            get
            {
                long sum = 0;
                for (int i = 0; i < this.valueBuckets.Length; ++i)
                {
                    sum += this.valueBuckets[i];
                }
                return sum;
            }
        }

        public HistogramThreadBuckets(int range, int valueBuckets) : base(range) 
        {
            this.valueBuckets = new long[valueBuckets];
        }

        public override void AddSample(int value)
        {
            ++sampleValueCounts[value];

            int bucket = Thread.CurrentThread.ManagedThreadId % this.valueBuckets.Length;
            ++valueBuckets[bucket];
        }
    }

    class HistogramThreadBucketsCacheLine : Histogram
    {
        public struct ValueBucket
        {
            private long _junk0;
            private long _junk1;
            private long _junk2;
            
            public long value;

            private long _junk4;
            private long _junk5;
            private long _junk6;
            private long _junk7;            
        };

        private ValueBucket[] valueBuckets;

        public ValueBucket[] ValueBuckets { get { return this.valueBuckets; } }

        public override long SampleCount
        {
            get
            {
                long sum = 0;
                for (int i = 0; i < this.valueBuckets.Length; ++i)
                {
                    sum += this.valueBuckets[i].value;
                }
                return sum;
            }
        }

        public HistogramThreadBucketsCacheLine(int range, int valueBuckets)
            : base(range) 
        {
            this.valueBuckets = new ValueBucket[valueBuckets];
        }

        public override void AddSample(int value, int threadId)
        {
            ++sampleValueCounts[value];

            ++valueBuckets[threadId].value;
        }
    }
}
